==> 0.5.0%2B7.tar.gz: lib/meta/meta_form.dart <==
import 'dart:async';

import 'package:after_layout/after_layout.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart' as provider;
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/meta/meta_form_context.dart';
import 'package:sunny_forms/meta/meta_form_control.dart';
import 'package:sunny_forms/meta/meta_form_service.dart';
import 'package:sunny_forms/meta/meta_property_handler.dart';
import 'package:sunny_forms/state_observer.dart';
import 'package:sunny_forms/sunny_fieldset.dart';
import 'package:sunny_forms/sunny_form/sunny_form_2.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_sdk_core/mverse/m_model.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/options.dart';

typedef MetaFormBuilderFn = Widget Function(
    BuildContext context, MetaFormState state);

typedef SubmitHook<T> = FutureOr<Result<T>> Function(BuildContext context,
    FormController form, FormSubmitResultBuilder submitResult);

SubmitHook<IFact?> saveFactCallback(MSchemaRef? ref) {
  return (BuildContext context, FormController form,
      FormSubmitResultBuilder result) async {
    if (form is MModelForm && form.mmodel is IFact) {
      final saved = await factService.saveFact(form.mmodel as IFact);
      return Result.success(saved);
    } else if (form is MModelForm) {
      var meetingRepo = apiRegistry.get(ref!);
      final saved = await meetingRepo.save(form.mmodel);
      return Result.success(saved as IFact);
    } else {
      return Result.noop();
    }
  };
}

/// Renders a form for a [IMSchemaDefinition] (based on [ref])
///
/// The form is backed by a [FormController], which is usually a [MModel/IFact] wrapped in a [MModelForm]
///
/// To override the behavior of the form, provide [MetaFormOverrides] through [MetaFormContext.overrides]
class MetaForm extends StatefulWidget {
  final String name;
  final MSchemaRef? ref;
  final FutureOr<MetaFormContext> metaFormContext;
  final List<MetaFormSection> sections;
  final MetaFormBuilderFn? builder;
  final FormControllerCreator? formCreator;
  final SubmitHook? submitHook;
  final FormValidator? validator;
  final bool? skipIfUnmodified;

  /// Whether this form should be attached as a subform to any parent form that exists in scope, either [parentForm]
  /// or a provided [Provider] form that's in scope.  Attaching to the parent form means that this form will validate
  /// and submit when the parent form submits.
  final bool attachForm;

  /// If inheritForm is true, then this [MetaForm] will bind directly to the parent form, instead of creating a new
  /// subform that's nested inside the parent form
  final bool inheritForm;
  final FormController? parentForm;

  final Future<Result> Function(
          BuildContext context, FormController form, FormSubmitResult result)?
      overrideHandler;

  MetaForm.ofMModel(
    BuildContext context, {

    /// A unique identifier for the form.  Typically, the form's key will be the id of the [model] instance, or
    /// the [ref], but sometimes you need a unique key, for example if you have two embedded add forms of the same type.
    String? id,

    /// If this form is a subform, this basePath will relocate
    /// all the child fields
    JsonPath? embedPath,
    required MModel? model,
    MSchemaRef? ref,
    MetadataOverrides? overrides,
    IMSchemaDefinition? schema,
    bool saveModel = true,
    this.parentForm,
    SubmitHook? submitHook,
    FormValidator? validator,

    /// Intercepts the submit handler, in case this form should submit its value somewhere other
    /// than the standard data endpoint
    this.overrideHandler,

    /// Whether this form should be attached as a subform to any parent form that exists in scope, either [parentForm]
    /// or a provided [Provider] form that's in scope.  Attaching to the parent form means that this form will validate
    /// and submit when the parent form submits.
    this.attachForm = true,

    /// If inheritForm is true, then this [MetaForm] will bind directly to the parent form, instead of creating a new
    /// subform that's nested inside the parent form
    this.inheritForm = false,
    Set? ignoredPaths,
    this.skipIfUnmodified,
    this.sections = const [],
    this.builder,
    List<MetaPropertyHandler> extraHandlers = const [],
    Map<String, dynamic>? scope,
  })  : assert(
            (id ?? (model is MEntity ? model.id : '') ?? ref ?? model?.mtype) !=
                null,
            "Must provide a valid type or identifier"),
        name = "${ref ?? model?.mtype}",
        ref = ref ?? model?.mtype,
        submitHook = inheritForm == true
            ? submitHook
            : submitHook ??
                (saveModel == true
                    ? ((overrideHandler ??
                        saveFactCallback(ref ?? model?.mtype)) as SubmitHook)
                    : null),
        formCreator = inheritForm == false
            ? ((context) => FormController.ofModel(context, model!))
            : null,
        validator = validator ??
            ((FormController form) async {
              final errors = await sunny
                  .get<IMSchemaService>(context: context)
                  .validate(model!);

              /// We validate the schema, but also check for formConfig violations
              final ov = overrides;
              ov?.formConfig?.forEach((path, value) {
                if (value.any((element) => element.isRequired == true)) {
                  final val = form.get(path);
                  if (val == null || (val is String && val.isNullOrBlank)) {
                    errors.add(ValidationError(
                        path: path, message: "You must enter a value"));
                  }
                }
              });
              return errors;
            }),
        metaFormContext = metaFormService.createContext(
          context,
          model,
          schemaRef: ref ?? model?.mtype,
          extraHandlers: extraHandlers,
          embedPath: embedPath,
          schema: schema,
          scope: scope,
          overrides: MetadataOverrides(ignoredPaths: {
                ...?ignoredPaths?.map((p) => JsonPath.of(p))
              }) +
              overrides,
        ),
        super(
            key: Key(
                "meta-form-${(id ?? (model is MEntity ? model.id : '') ?? ref ?? model?.mtype)}"));

  MetaForm({
    required this.name,
    required this.metaFormContext,
    BuildContext? context,
    MSchemaRef? ref,

    /// Intercepts the submit handler, in case this form should submit its value somewhere other
    /// than the standard data endpoint
    this.overrideHandler,
    this.builder,
    this.submitHook,
    this.validator,
    this.inheritForm = false,
    this.parentForm,
    this.formCreator,
    this.attachForm = true,
    this.skipIfUnmodified,
    this.sections = const [],
  })  : ref = ref ??
            metaFormContext.resolveOrNull()?.ref ??
            illegalState("Must provide a non-future metaFormContext or a ref"),
        super(key: Key("meta-form-$name"));

  @override
  State<StatefulWidget> createState() => MetaFormState();
}

class MetaFormState extends StateObserver<MetaForm> with AfterLayoutMixin {
  MetaFormState();

  FocusNode? focusNode;
  final focusNodes = <String?, FocusNode>{};
  late Map<JsonPath, IMSchemaProperty> _properties;
  late List<MetaFormSection> _sections;
  final GlobalKey<FormState> formStateKey = GlobalKey<FormState>();

  MetaFormContext? _metaForm;
  IMSchemaDefinition? _schema;

  IMSchemaDefinition? get schema => _schema;

  bool get isEmbed => _metaForm!.embedPath != null;

  MetadataOverrides get _metaOverrides => _metaForm!.overrides;

  Future initializeFromSchema(IMSchemaDefinition schema) async {
    if (this._schema != null) {
      log.info("Already initialized");
      return;
    }

    this._schema ??= schema;

    _properties = mapOf<JsonPath, IMSchemaProperty>(
      _schema!.properties.embedded(_metaForm!.embedPath)!,
      keyOf: (prop) => prop.jsonPath,
    );

    _properties.values.forEach((prop) {
      final _fn = FocusNode();
      if (focusNode == null && prop.isRequired!) {
        focusNode = _fn;
      }
      focusNodes[prop.path] = _fn;
    });

    // Maps property path to existing section definition
    final pathToSection = Map.fromEntries(
      widget.sections.expand(
        (section) => section.properties.map((prop) => MapEntry(prop, section)),
      ),
    );

    _sections = [
      ...widget.sections,
      MetaFormSection(
          key: "default",
          properties: _metaOverrides
              .sortProperties(_metaForm!.handledPaths.where((p) {
                return !pathToSection.containsKey(p);
              }))
              .toSet()),
    ];

    focusNode ??= focusNodes.entries.firstOrNull?.value;
  }

  bool initController = false;

  @override
  void initState() {
    super.initState();

    asyncInit();
  }

  Future asyncInit() async {
    _metaForm = await widget.metaFormContext;
    final mschemaDefinition = await _metaForm!.loaded;
    await initializeFromSchema(mschemaDefinition);
    safeState(() {});
  }

  @override
  void dispose() {
    super.dispose();
  }

  /// Used to register with parent form
  String get subformKey => "${widget.ref}";

  @override
  Widget build(BuildContext context) {
    return _metaForm == null
        ? loadingIndicator
        : provider.MultiProvider(
            key: Key("multi-providers-${widget.ref}"),
            providers: [
              provider.Provider<IMetaFormContext>.value(
                  value: _metaForm!.withOverrides(context.find())),
              provider.Provider<FormController?>(
                key: Key("meta-form-controller-${widget.ref}"),
                create: (context) {
                  /// Used to register with the parent form
                  FormController? form;

                  if (widget.formCreator != null) {
                    form = widget.formCreator?.call(context);

                    if (widget.parentForm != null) {
                      widget.parentForm!.registerForm(form);
                    } else if (widget.attachForm != false) {
                      context.formOrNull(listen: false)?.also((parent) {
                        parent.registerForm(form);
                      });
                    }
                  } else {
                    form = context.form();
                  }

                  if (form!.parent == null) {
                    /// If we have a parent, we shouldn't try to also bind - it might overwrite the existing binding
                    form.tryBindCommand(context,
                        skipIfUnmodified: widget.skipIfUnmodified);
                  }

                  if (widget.submitHook != null) {
                    form.registerSubmitHook(subformKey, (form, result) {
                      return widget.submitHook!(context, form, result);
                    });
                  }
                  if (widget.validator != null) {
                    form.registerFormValidator(subformKey, widget.validator);
                  }
                  return form;
                },
                dispose: (context, form) {
                  if (widget.formCreator != null) {
                    form?.dispose();
                  }
                  form?.unregisterSubmitHook(subformKey);
                  form?.unregisterFormValidator(subformKey);
                },
              ),
            ],
            child: provider.Consumer2<FormController, IMetaFormContext>(
              builder:
                  (BuildContext context, controller, mfContext, Widget? child) {
                return _schema == null
                    ? emptyBox
                    : widget.builder?.call(context, this) ??
                        _defaultBuilder(context,
                            name: widget.name,
                            form: controller,
                            mfContext: mfContext,
                            isEmbed: isEmbed,
                            sections: _sections);
                // return crossFade(
                //   first: _schema == null
                //       ? emptyBox
                //       : widget.builder?.call(context, this) ??
                //           _defaultBuilder(context,
                //               name: widget.name,
                //               form: controller,
                //               mfContext: mfContext,
                //               isEmbed: isEmbed,
                //               sections: _sections),
                //   second: SunnyPad(child: Text("Loading...")),
                //   showFirst: _schema != null,
                // );
              },
            ),
          );
  }

  @override
  void afterFirstLayout(BuildContext context) {
    if (focusNode != null) {
      FocusScope.of(context).autofocus(focusNode!);
    }
  }

  @override
  setState(VoidCallback fn) {
    super.setState(fn);
  }

  static Widget _defaultBuilder(BuildContext context,
      {required String name,
      required FormController form,
      required IMetaFormContext mfContext,
      required bool isEmbed,
      required List<MetaFormSection> sections}) {
    final _metaForm = mfContext as MetaFormContext;

    SunnyFormField? buildField(JsonPath fieldPath) {
      return _metaForm.overrides.ignoredPaths!.contains(fieldPath)
          ? null
          : SunnyFormField.meta(
              context,
              key: null,
              formMetadata: _metaForm,
              attribute: fieldPath,
            );
    }

    // ignore: deprecated_member_use_from_same_package
    return SunnyForm2(
      name,
      form: form,
      embed: isEmbed,
      fieldsets: sections.map(
        (MetaFormSection section) {
          if (section.isEmpty) {
            return null;
          }

          final sectionProps = section.properties
              .orEmpty()
              .map(buildField)
              .whereNotNull()
              .toList() as List<SunnyFormField<dynamic>>;

          return SunnyFieldset.of(
            embed: isEmbed,
            children: [
              ...sectionProps,
            ],
          );
        },
      ).notNull(),
    );
  }
}

// final isoDateFormat = DateFormat("yyyy-MM-dd");

extension MSchemaPropertyRelocationExtension on IMSchemaProperty {
  IMSchemaProperty embedded(JsonPath basePath) {
    final existingPath = jsonPath;
    return IMSchemaProperty.of(
      type: type,
      path: "${basePath + existingPath}",
      label: label,
      description: description,
      uri: uri,
      baseCode: baseCode,
      isRequired: isRequired,
    );
  }
}

extension LListOfMSchemaPropertyRelocationExtension on List<IMSchemaProperty>? {
  List<IMSchemaProperty>? embedded(JsonPath? basePath) {
    if (basePath == null || basePath == JsonPath.root()) {
      return this;
    }
    return [...?this?.map((_) => _.embedded(basePath))];
  }
}

void _ensureWiring() {
  if (IFormController.defaultMModelControllerFactory == null) {
    IFormController.defaultMModelControllerFactory = (context, model) {
      return FormController.ofModel<MModel>(context, model);
    };
  }
}

extension MModelMetaFormExt on MModel {
  Widget buildForm(
    BuildContext context, {
    String? id,
  }) {
    _ensureWiring();
    final ref = this.mtype;
    return mschemaService.getSchema(ref).build((schema) => MetaForm.ofMModel(
          context,
          model: this,
          ref: ref,
          id: id,
        ));
  }
}

extension MSchemaRefFormExt on MSchemaRef {
  Widget buildForm(
    BuildContext context, {
    FormController? form,
    MModel? record,
    MetadataOverrides? overrides,
  }) {
    _ensureWiring();
    final ref = this;
    record ??= mmodelRegistry.instantiate(json: <String, dynamic>{}, type: ref);
    return metaFormService.buildForm(context, ref,
        controller: form, overrides: overrides, record: record);
  }
}

==> 0.6.0%2B7.tar.gz: lib/meta/meta_form.dart <==
import 'dart:async';

import 'package:after_layout/after_layout.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart' as provider;
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/meta/meta_form_context.dart';
import 'package:sunny_forms/meta/meta_form_control.dart';
import 'package:sunny_forms/meta/meta_form_service.dart';
import 'package:sunny_forms/meta/meta_property_handler.dart';
import 'package:sunny_forms/sunny_fieldset.dart';
import 'package:sunny_forms/sunny_form/sunny_form_2.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_sdk_core/mverse/m_base_model.dart';
import 'package:sunny_sdk_core/mverse/m_model.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/options.dart';

typedef MetaFormBuilderFn = Widget Function(BuildContext context, MetaFormState state);

typedef SubmitHook<T> = FutureOr<Result<T>> Function(
    BuildContext context, FormController form, FormSubmitResultBuilder submitResult);

SubmitHook<IFact?> saveFactCallback(MSchemaRef? ref) {
  return (BuildContext context, FormController form, FormSubmitResultBuilder result) async {
    if (form is MModelForm && form.mmodel is IFact) {
      final saved = await factService.saveFact(form.mmodel as IFact);
      return Result.success(saved);
    } else if (form is MModelForm && form.mmodel is MModel) {
      var repo = apiRegistry.get(ref!);
      final saved = await repo.save(form.mmodel as MModel);
      return Result.success(saved as IFact);
    } else {
      return Result.noop();
    }
  };
}

/// Renders a form for a [IMSchemaDefinition] (based on [ref])
///
/// The form is backed by a [FormController], which is usually a [MModel/IFact] wrapped in a [MModelForm]
///
/// To override the behavior of the form, provide [MetaFormOverrides] through [MetaFormContext.overrides]
class MetaForm extends StatefulWidget {
  final String name;
  final MSchemaRef? ref;
  final FutureOr<MetaFormContext> metaFormContext;
  final List<MetaFormSection> sections;
  final MetaFormBuilderFn? builder;
  final FormControllerCreator? formCreator;
  final SubmitHook? submitHook;
  final FormValidator? validator;
  final bool? skipIfUnmodified;

  /// Whether this form should be attached as a subform to any parent form that exists in scope, either [parentForm]
  /// or a provided [Provider] form that's in scope.  Attaching to the parent form means that this form will validate
  /// and submit when the parent form submits.
  final bool attachForm;

  /// If inheritForm is true, then this [MetaForm] will bind directly to the parent form, instead of creating a new
  /// subform that's nested inside the parent form
  final bool inheritForm;
  final FormController? parentForm;

  final Future<Result> Function(BuildContext context, FormController form, FormSubmitResult result)? overrideHandler;

  MetaForm.ofMModel(
    BuildContext context, {

    /// A unique identifier for the form.  Typically, the form's key will be the id of the [model] instance, or
    /// the [ref], but sometimes you need a unique key, for example if you have two embedded add forms of the same type.
    String? id,

    /// If this form is a subform, this basePath will relocate
    /// all the child fields
    JsonPath? embedPath,
    required MBaseModel? model,
    MSchemaRef? ref,
    MetadataOverrides? overrides,
    IMSchemaDefinition? schema,
    bool saveModel = true,
    this.parentForm,
    SubmitHook? submitHook,
    FormValidator? validator,

    /// Intercepts the submit handler, in case this form should submit its value somewhere other
    /// than the standard data endpoint
    this.overrideHandler,

    /// Whether this form should be attached as a subform to any parent form that exists in scope, either [parentForm]
    /// or a provided [Provider] form that's in scope.  Attaching to the parent form means that this form will validate
    /// and submit when the parent form submits.
    this.attachForm = true,

    /// If inheritForm is true, then this [MetaForm] will bind directly to the parent form, instead of creating a new
    /// subform that's nested inside the parent form
    this.inheritForm = false,
    Set? ignoredPaths,
    this.skipIfUnmodified,
    this.sections = const [],
    this.builder,
    List<MetaPropertyHandler> extraHandlers = const [],
    Map<String, dynamic>? scope,
  })  : assert(
            (id ?? (model is MEntity ? model.id : '') ?? ref ?? model?.mtype) != null, "Must provide a valid type or identifier"),
        name = "${ref ?? model?.mtype}",
        ref = ref ?? model?.mtype,
        submitHook = inheritForm == true
            ? submitHook
            : submitHook ??
                (saveModel == true ? ((overrideHandler ?? saveFactCallback(ref ?? model?.mtype)) as SubmitHook) : null),
        formCreator = inheritForm == false ? ((context) => FormController.ofModel(context, model!)) : null,
        validator = validator ??
            ((FormController form) async {
              final errors =
                  model is MModel ? (await sunny.get<IMSchemaService>(context: context).validate(model)) : <ValidationError>[];

              /// We validate the schema, but also check for formConfig violations
              final ov = overrides;
              ov?.formConfig?.forEach((path, value) {
                if (value.any((element) => element.isRequired == true)) {
                  final val = form.get(path);
                  if (val == null || (val is String && val.isNullOrBlank)) {
                    errors.add(ValidationError(path: path, message: "You must enter a value"));
                  }
                }
              });
              return errors;
            }),
        metaFormContext = metaFormService.createContext(
          context,
          model,
          schemaRef: ref ?? model?.mtype,
          extraHandlers: extraHandlers,
          embedPath: embedPath,
          schema: schema,
          scope: scope,
          overrides: MetadataOverrides(ignoredPaths: {...?ignoredPaths?.map((p) => JsonPath.of(p))}) + overrides,
        ),
        super(key: Key("meta-form-${(id ?? (model is MEntity ? model.id : '') ?? ref ?? model?.mtype)}"));

  MetaForm({
    required this.name,
    required this.metaFormContext,
    BuildContext? context,
    MSchemaRef? ref,

    /// Intercepts the submit handler, in case this form should submit its value somewhere other
    /// than the standard data endpoint
    this.overrideHandler,
    this.builder,
    this.submitHook,
    this.validator,
    this.inheritForm = false,
    this.parentForm,
    this.formCreator,
    this.attachForm = true,
    this.skipIfUnmodified,
    this.sections = const [],
  })  : ref = ref ?? metaFormContext.resolveOrNull()?.ref ?? illegalState("Must provide a non-future metaFormContext or a ref"),
        super(key: Key("meta-form-$name"));

  @override
  State<StatefulWidget> createState() => MetaFormState();
}

class MetaFormState extends BaseState<MetaForm> with AfterLayoutMixin {
  MetaFormState();

  FocusNode? focusNode;
  final focusNodes = <String?, FocusNode>{};
  late Map<JsonPath, IMSchemaProperty> _properties;
  late List<MetaFormSection> _sections;
  final GlobalKey<FormState> formStateKey = GlobalKey<FormState>();

  MetaFormContext? _metaForm;
  IMSchemaDefinition? _schema;

  IMSchemaDefinition? get schema => _schema;

  bool get isEmbed => _metaForm!.embedPath != null;

  MetadataOverrides get _metaOverrides => _metaForm!.overrides;

  Future initializeFromSchema(IMSchemaDefinition schema) async {
    if (this._schema != null) {
      log.info("Already initialized");
      return;
    }

    this._schema ??= schema;

    _properties = mapOf<JsonPath, IMSchemaProperty>(
      _schema!.properties.embedded(_metaForm!.embedPath)!,
      keyOf: (prop) => prop.jsonPath,
    );

    _properties.values.forEach((prop) {
      final _fn = FocusNode();
      if (focusNode == null && prop.isRequired!) {
        focusNode = _fn;
      }
      focusNodes[prop.path] = _fn;
    });

    // Maps property path to existing section definition
    final pathToSection = Map.fromEntries(
      widget.sections.expand(
        (section) => section.properties.map((prop) => MapEntry(prop, section)),
      ),
    );

    _sections = [
      ...widget.sections,
      MetaFormSection(
          key: "default",
          properties: _metaOverrides
              .sortProperties(_metaForm!.handledPaths.where((p) {
                return !pathToSection.containsKey(p);
              }))
              .toSet()),
    ];

    focusNode ??= focusNodes.entries.firstOrNull?.value;
  }

  bool initController = false;

  @override
  void initState() {
    super.initState();

    asyncInit();
  }

  Future asyncInit() async {
    _metaForm = await widget.metaFormContext;
    final mschemaDefinition = await _metaForm!.loaded;
    await initializeFromSchema(mschemaDefinition);
    safeState(() {});
  }

  @override
  void dispose() {
    super.dispose();
  }

  /// Used to register with parent form
  String get subformKey => "${widget.ref}";

  @override
  Widget build(BuildContext context) {
    return _metaForm == null
        ? loadingIndicator
        : provider.MultiProvider(
            key: Key("multi-providers-${widget.ref}"),
            providers: [
              provider.Provider<IMetaFormContext>.value(value: _metaForm!.withOverrides(context.find())),
              provider.Provider<FormController?>(
                key: Key("meta-form-controller-${widget.ref}"),
                create: (context) {
                  /// Used to register with the parent form
                  FormController? form;

                  if (widget.formCreator != null) {
                    form = widget.formCreator?.call(context);

                    if (widget.parentForm != null) {
                      widget.parentForm!.registerForm(form);
                    } else if (widget.attachForm != false) {
                      context.formOrNull(listen: false)?.also((parent) {
                        parent.registerForm(form);
                      });
                    }
                  } else {
                    form = context.form();
                  }

                  if (form!.parent == null) {
                    /// If we have a parent, we shouldn't try to also bind - it might overwrite the existing binding
                    form.tryBindCommand(context, skipIfUnmodified: widget.skipIfUnmodified);
                  }

                  if (widget.submitHook != null) {
                    form.registerSubmitHook(subformKey, (form, result) {
                      return widget.submitHook!(context, form, result);
                    });
                  }
                  if (widget.validator != null) {
                    form.registerFormValidator(subformKey, widget.validator);
                  }
                  return form;
                },
                dispose: (context, form) {
                  if (widget.formCreator != null) {
                    form?.dispose();
                  }
                  form?.unregisterSubmitHook(subformKey);
                  form?.unregisterFormValidator(subformKey);
                },
              ),
            ],
            child: provider.Consumer2<FormController, IMetaFormContext>(
              builder: (BuildContext context, controller, mfContext, Widget? child) {
                return _schema == null
                    ? emptyBox
                    : widget.builder?.call(context, this) ??
                        _defaultBuilder(context,
                            name: widget.name, form: controller, mfContext: mfContext, isEmbed: isEmbed, sections: _sections);
                // return crossFade(
                //   first: _schema == null
                //       ? emptyBox
                //       : widget.builder?.call(context, this) ??
                //           _defaultBuilder(context,
                //               name: widget.name,
                //               form: controller,
                //               mfContext: mfContext,
                //               isEmbed: isEmbed,
                //               sections: _sections),
                //   second: SunnyPad(child: Text("Loading...")),
                //   showFirst: _schema != null,
                // );
              },
            ),
          );
  }

  @override
  void afterFirstLayout(BuildContext context) {
    if (focusNode != null) {
      FocusScope.of(context).autofocus(focusNode!);
    }
  }

  @override
  setState(VoidCallback fn) {
    super.setState(fn);
  }

  static Widget _defaultBuilder(BuildContext context,
      {required String name,
      required FormController form,
      required IMetaFormContext mfContext,
      required bool isEmbed,
      required List<MetaFormSection> sections}) {
    final _metaForm = mfContext as MetaFormContext;

    SunnyFormField? buildField(JsonPath fieldPath) {
      return _metaForm.overrides.ignoredPaths!.contains(fieldPath)
          ? null
          : SunnyFormField.meta(
              context,
              key: null,
              formMetadata: _metaForm,
              attribute: fieldPath,
            );
    }

    // ignore: deprecated_member_use_from_same_package
    return SunnyForm2(
      name,
      form: form,
      embed: isEmbed,
      fieldsets: sections.map(
        (MetaFormSection section) {
          if (section.isEmpty) {
            return null;
          }

          final sectionProps = section.properties.orEmpty().mapNotNull(buildField).toList();

          return SunnyFieldset.of(
            embed: isEmbed,
            children: [
              ...sectionProps,
            ],
          );
        },
      ).notNull(),
    );
  }
}

// final isoDateFormat = DateFormat("yyyy-MM-dd");

extension MSchemaPropertyRelocationExtension on IMSchemaProperty {
  IMSchemaProperty embedded(JsonPath basePath) {
    final existingPath = jsonPath;
    return IMSchemaProperty.of(
      type: type,
      path: "${basePath + existingPath}",
      label: label,
      description: description,
      uri: uri,
      baseCode: baseCode,
      isRequired: isRequired,
    );
  }
}

extension LListOfMSchemaPropertyRelocationExtension on List<IMSchemaProperty>? {
  List<IMSchemaProperty>? embedded(JsonPath? basePath) {
    if (basePath == null || basePath == JsonPath.root()) {
      return this;
    }
    return [...?this?.map((_) => _.embedded(basePath))];
  }
}

void _ensureWiring() {
  if (IFormController.defaultMModelControllerFactory == null) {
    IFormController.defaultMModelControllerFactory = (context, model) {
      return FormController.ofModel<MModel>(context, model);
    };
  }
}

extension MModelMetaFormExt on MModel {
  Widget buildForm(
    BuildContext context, {
    String? id,
  }) {
    _ensureWiring();
    final ref = this.mtype;
    return mschemaService.getSchema(ref).build((schema) => MetaForm.ofMModel(
          context,
          model: this,
          ref: ref,
          id: id,
        ));
  }
}

extension MSchemaRefFormExt on MSchemaRef {
  Widget buildForm(
    BuildContext context, {
    FormController? form,
    MBaseModel? record,
    MetadataOverrides? overrides,
  }) {
    _ensureWiring();
    final ref = this;
    record ??= mmodelRegistry.instantiate(json: <String, dynamic>{}, type: ref);
    return context.metaFormService.buildForm(context, ref, controller: form, overrides: overrides, record: record);
  }
}
